% MOMGRAPH   Christoffel-Darboux semialgebraic approximation of a graph from its moments
%
% Given a function f(x) from R^{p-1} to [-1,1] and the grevlex monomial basis b(x,y)
% define its moment matrix as M = \int b(x,f(x))b(x,f(x))'dx
%
% Given M and a collection of points X = [x1; x2; .. xN], the command
%
%   Y = MOMGRAPH(M,X)
%
% returns an approximation Y = [y1; y2; .. yN] of the values [f(x1); f(x2); .. f(xN)]
% obtained from the regularized Christoffel-Darboux polynomial of M
%
% With additional input argument regpar, the command
%
%  [Y,P] = MOMGRAPH(M,X,regpar)
%
% also returns a matrix P of coefficients such that the Christoffel-Darboux
% polynomial with regularization parameter regpar is the sum of squares b'*P'*P*b
%
% Input format:
% - M symmetric numerical matrix (class DOUBLE) of size n by n
% - X numerical matrix (class DOUBLE) of size N by p-1
% - regpar positive real number (class DOUBLE), default value 1e-8
%
% Output format:
% - Y numerical vector (class DOUBLE) of size N by 1
% - P numerical matrix (class DOUBLE) of size N by N

% Requires the external function POWVEC.M
%
% S. Marx, E. Pauwels, T. Weisser, D. Henrion, J. B. Lasserre, March 2019

function [Y,P] = momgraph(M,X,regpar)

if nargin<3, regpar = 1e-8; end
[Nx,p1] = size(X); p = p1+1;
d=1; while nchoosek(p+d,p)<length(M), d=d+1; end

% eigenvalue decomposition of the moment matrix
[eigvec,eigval] = eig(M);
e = diag(eigval)+regpar;

% construct Christoffel-Darboux coefficient matrix
if nargout>1, P = diag(e.^(-1/2))*eigvec'; end

Ny = 1e3; % number of samples in y
Y = linspace(-1,1,Ny);
samples = [repmat(X,Ny,1),repelem(Y,Nx)']';

% evaluate monomials at the samples
powers = powvec(p+1,d); powers(:,1)=[];
monomials = exp(powers*log(abs(samples)+eps));
flipsign = find(rem(powers*(1-sign(samples))/2,2));
monomials(flipsign) = -monomials(flipsign);

% evaluate Christoffel-Darboux polynomial at the samples
q = e'.^(-1)*(eigvec'*monomials).^2;

% minimize y for each x
[~,indices] = min(reshape(q',Nx,Ny),[],2);
Y = Y(indices);
